@@ -1,5 +1,5 @@
 /* $RoughId: md5init.c,v 1.2 2001/07/13 19:49:10 knu Exp $ */
-/* $Id: md5init.c 25189 2009-10-02 12:04:37Z akr $ */
+/* $Id$ */
 
 #include "digest.h"
 #if defined(HAVE_OPENSSL_MD5_H)
@@ -24,17 +24,22 @@ static const rb_digest_metadata_t md5 = {
  * RFC1321.
  */
 void
-Init_md5()
+Init_md5(void)
 {
     VALUE mDigest, cDigest_Base, cDigest_MD5;
 
     rb_require("digest");
 
+#if 0
+    mDigest = rb_define_module("Digest"); /* let rdoc know */
+#endif
     mDigest = rb_path2class("Digest");
     cDigest_Base = rb_path2class("Digest::Base");
 
     cDigest_MD5 = rb_define_class_under(mDigest, "MD5", cDigest_Base);
 
+#undef RUBY_UNTYPED_DATA_WARNING
+#define RUBY_UNTYPED_DATA_WARNING 0
     rb_ivar_set(cDigest_MD5, rb_intern("metadata"),
-      Data_Wrap_Struct(rb_cObject, 0, 0, (void *)&md5));
+		Data_Wrap_Struct(0, 0, 0, (void *)&md5));
 }
