@@ -1,8 +1,7 @@
 /* $RoughId: rmd160init.c,v 1.3 2001/07/13 20:00:43 knu Exp $ */
-/* $Id: rmd160init.c 52797 2015-11-30 12:08:16Z usa $ */
+/* $Id$ */
 
-#include <ruby/ruby.h>
-#include "../digest.h"
+#include "digest.h"
 #if defined(HAVE_OPENSSL_RIPEMD_H)
 #include "rmd160ossl.h"
 #else
@@ -25,7 +24,7 @@ static const rb_digest_metadata_t rmd160 = {
  * Bosselaers, and Bart Preneel.
  */
 void
-Init_rmd160()
+Init_rmd160(void)
 {
     VALUE mDigest, cDigest_Base, cDigest_RMD160;
 
@@ -39,6 +38,8 @@ Init_rmd160()
 
     cDigest_RMD160 = rb_define_class_under(mDigest, "RMD160", cDigest_Base);
 
+#undef RUBY_UNTYPED_DATA_WARNING
+#define RUBY_UNTYPED_DATA_WARNING 0
     rb_ivar_set(cDigest_RMD160, rb_intern("metadata"),
-      Data_Wrap_Struct(rb_cObject, 0, 0, (void *)&rmd160));
+		Data_Wrap_Struct(0, 0, 0, (void *)&rmd160));
 }
