@@ -1,5 +1,6 @@
+# -*- coding: us-ascii -*-
 # $RoughId: extconf.rb,v 1.3 2001/08/14 19:54:51 knu Exp $
-# $Id: extconf.rb 25189 2009-10-02 12:04:37Z akr $
+# $Id$
 
 require "mkmf"
 
@@ -8,10 +9,12 @@ $INCFLAGS << " -I$(srcdir)/.."
 
 $objs = [ "sha1init.#{$OBJEXT}" ]
 
-dir_config("openssl")
-
 if !with_config("bundled-sha1") &&
-    have_library("crypto") && have_header("openssl/sha.h")
+    (dir_config("openssl")
+     pkg_config("openssl")
+     require File.expand_path('../../../openssl/deprecation', __FILE__)
+     have_library("crypto")) &&
+    OpenSSL.check_func("SHA1_Transform", "openssl/sha.h")
   $objs << "sha1ossl.#{$OBJEXT}"
 else
   $objs << "sha1.#{$OBJEXT}"
