@@ -1,6 +1,6 @@
 /*	$NetBSD: sha1.h,v 1.2 1998/05/29 22:55:44 thorpej Exp $	*/
 /*	$RoughId: sha1.h,v 1.3 2002/02/24 08:14:32 knu Exp $	*/
-/*	$Id: sha1.h 27437 2010-04-22 08:04:13Z nobu $	*/
+/*	$Id$	*/
 
 /*
  * SHA-1 in C
@@ -28,9 +28,9 @@ typedef struct {
 #endif
 
 void	SHA1_Transform _((uint32_t state[5], const uint8_t buffer[64]));
-void	SHA1_Init _((SHA1_CTX *context));
+int	SHA1_Init _((SHA1_CTX *context));
 void	SHA1_Update _((SHA1_CTX *context, const uint8_t *data, size_t len));
-void	SHA1_Finish _((SHA1_CTX *context, uint8_t digest[20]));
+int	SHA1_Finish _((SHA1_CTX *context, uint8_t digest[20]));
 
 #define SHA1_BLOCK_LENGTH		64
 #define SHA1_DIGEST_LENGTH		20
