@@ -1,5 +1,6 @@
+# -*- coding: us-ascii -*-
 # $RoughId: extconf.rb,v 1.4 2001/08/14 19:54:51 knu Exp $
-# $Id: extconf.rb 35653 2012-05-15 16:10:46Z tenderlove $
+# $Id$
 
 require "mkmf"
 
@@ -8,12 +9,12 @@ $INCFLAGS << " -I$(srcdir)/.."
 
 $objs = [ "sha2init.#{$OBJEXT}" ]
 
-dir_config("openssl")
-pkg_config("openssl")
-
 if !with_config("bundled-sha2") &&
-    have_library("crypto") &&
-    %w[SHA256 SHA512].all? {|d| have_func("#{d}_Transform", "openssl/sha.h")} &&
+    (dir_config("openssl")
+     pkg_config("openssl")
+     require File.expand_path('../../../openssl/deprecation', __FILE__)
+     have_library("crypto")) &&
+    %w[SHA256 SHA512].all? {|d| OpenSSL.check_func("#{d}_Transform", "openssl/sha.h")} &&
     %w[SHA256 SHA512].all? {|d| have_type("#{d}_CTX", "openssl/sha.h")}
   $objs << "sha2ossl.#{$OBJEXT}"
   $defs << "-DSHA2_USE_OPENSSL"
