@@ -2,7 +2,7 @@
 
   tkutil.c -
 
-  $Author: nobu $
+  $Author$
   created at: Fri Nov  3 00:47:54 JST 1995
 
 ************************************************/
@@ -23,6 +23,9 @@ static int rb_thread_critical; /* dummy */
 #include "st.h"
 #endif
 
+#undef RUBY_UNTYPED_DATA_WARNING
+#define RUBY_UNTYPED_DATA_WARNING 1
+
 #if !defined(RHASH_TBL)
 #define RHASH_TBL(h) (RHASH(h)->tbl)
 #endif
@@ -312,7 +315,7 @@ ary2list(ary, enc_flag, self)
     if (NIL_P(enc_flag)) {
         dst_enc = sys_enc;
         req_chk_flag = 1;
-    } else if (TYPE(enc_flag) == T_TRUE || TYPE(enc_flag) == T_FALSE) {
+    } else if (enc_flag == Qtrue || enc_flag == Qfalse) {
         dst_enc = enc_flag;
         req_chk_flag = 0;
     } else {
@@ -323,7 +326,7 @@ ary2list(ary, enc_flag, self)
     /* size = RARRAY_LEN(ary); */
     size = 0;
     for(idx = 0; idx < RARRAY_LEN(ary); idx++) {
-        if (TYPE(RARRAY_PTR(ary)[idx]) == T_HASH) {
+        if (RB_TYPE_P(RARRAY_PTR(ary)[idx], T_HASH)) {
             size += 2 * RHASH_SIZE(RARRAY_PTR(ary)[idx]);
         } else {
             size++;
@@ -433,7 +436,7 @@ ary2list(ary, enc_flag, self)
             RARRAY_PTR(dst)[idx] = str_val;
         }
         val = rb_apply(cTclTkLib, ID_merge_tklist, dst);
-        if (TYPE(dst_enc) == T_STRING) {
+        if (RB_TYPE_P(dst_enc, T_STRING)) {
             val = rb_funcall(cTclTkLib, ID_fromUTF8, 2, val, dst_enc);
             rb_ivar_set(val, ID_at_enc, dst_enc);
         } else {
@@ -466,7 +469,7 @@ ary2list2(ary, enc_flag, self)
     if (NIL_P(enc_flag)) {
         dst_enc = sys_enc;
         req_chk_flag = 1;
-    } else if (TYPE(enc_flag) == T_TRUE || TYPE(enc_flag) == T_FALSE) {
+    } else if (enc_flag == Qtrue || enc_flag == Qfalse) {
         dst_enc = enc_flag;
         req_chk_flag = 0;
     } else {
@@ -527,7 +530,7 @@ ary2list2(ary, enc_flag, self)
             RARRAY_PTR(dst)[idx] = str_val;
         }
         val = rb_apply(cTclTkLib, ID_merge_tklist, dst);
-        if (TYPE(dst_enc) == T_STRING) {
+        if (RB_TYPE_P(dst_enc, T_STRING)) {
             val = rb_funcall(cTclTkLib, ID_fromUTF8, 2, val, dst_enc);
             rb_ivar_set(val, ID_at_enc, dst_enc);
         } else {
@@ -561,7 +564,7 @@ assoc2kv(assoc, ary, self)
 
     for(i = 0; i < len; i++) {
         pair = RARRAY_PTR(assoc)[i];
-        if (TYPE(pair) != T_ARRAY) {
+        if (!RB_TYPE_P(pair, T_ARRAY)) {
             rb_ary_push(dst, key2keyname(pair));
             continue;
         }
@@ -609,7 +612,7 @@ assoc2kv_enc(assoc, ary, self)
 
     for(i = 0; i < len; i++) {
         pair = RARRAY_PTR(assoc)[i];
-        if (TYPE(pair) != T_ARRAY) {
+        if (!RB_TYPE_P(pair, T_ARRAY)) {
             rb_ary_push(dst, key2keyname(pair));
             continue;
         }
@@ -831,15 +834,15 @@ get_eval_string_core(obj, enc_flag, self)
         if (RTEST(enc_flag)) {
             if (rb_obj_respond_to(self, ID_toUTF8, Qtrue)) {
                 return rb_funcall(self, ID_toUTF8, 1,
-                                  rb_str_new2(rb_id2name(SYM2ID(obj))));
+                                  rb_str_dup(rb_sym2str(obj)));
             } else {
-                return fromDefaultEnc_toUTF8(rb_str_new2(rb_id2name(SYM2ID(obj))), self);
+                return fromDefaultEnc_toUTF8(rb_sym2str(obj), self);
             }
         } else {
 #ifdef HAVE_RB_SYM_TO_S
             return rb_sym_to_s(obj);
 #else
-            return rb_str_new2(rb_id2name(SYM2ID(obj)));
+            return rb_sym2str(obj);
 #endif
         }
 
@@ -950,7 +953,7 @@ tk_conv_args(argc, argv, self)
     old_gc = rb_gc_disable();
 
     for(size = 0, idx = 2; idx < argc; idx++) {
-        if (TYPE(argv[idx]) == T_HASH) {
+        if (RB_TYPE_P(argv[idx], T_HASH)) {
             size += 2 * RHASH_SIZE(argv[idx]);
         } else {
             size++;
@@ -959,7 +962,7 @@ tk_conv_args(argc, argv, self)
     /* dst = rb_ary_new2(argc - 2); */
     dst = rb_ary_new2(size);
     for(idx = 2; idx < argc; idx++) {
-        if (TYPE(argv[idx]) == T_HASH) {
+        if (RB_TYPE_P(argv[idx], T_HASH)) {
             if (RTEST(argv[1])) {
                 hash2kv_enc(argv[idx], dst, self);
             } else {
@@ -984,7 +987,7 @@ tcl2rb_bool(self, value)
     VALUE self;
     VALUE value;
 {
-    if (TYPE(value) == T_FIXNUM) {
+    if (RB_TYPE_P(value, T_FIXNUM)) {
         if (NUM2INT(value) == 0) {
             return Qfalse;
         } else {
@@ -992,7 +995,7 @@ tcl2rb_bool(self, value)
         }
     }
 
-    if (TYPE(value) == T_TRUE || TYPE(value) == T_FALSE) {
+    if (value == Qtrue || value == Qfalse) {
         return value;
     }
 
@@ -1168,6 +1171,23 @@ subst_free(ptr)
     }
 }
 
+static size_t
+subst_memsize(ptr)
+    const struct cbsubst_info *ptr;
+{
+    return sizeof(*ptr);
+}
+
+static const rb_data_type_t cbsubst_info_type = {
+    "TkUtil/CallbackSubst/Info",
+    {
+	subst_mark,
+	subst_free,
+	subst_memsize,
+    },
+    0, 0, RUBY_TYPED_FREE_IMMEDIATELY,
+};
+
 static VALUE
 allocate_cbsubst_info(struct cbsubst_info **inf_ptr)
 {
@@ -1194,16 +1214,23 @@ allocate_cbsubst_info(struct cbsubst_info **inf_ptr)
 
   if (inf_ptr != (struct cbsubst_info **)NULL) *inf_ptr = inf;
 
-  return Data_Wrap_Struct(cSUBST_INFO, subst_mark, subst_free, inf);
+  return TypedData_Wrap_Struct(cSUBST_INFO, &cbsubst_info_type, inf);
 }
 
 static void
-cbsubst_init()
+cbsubst_init(void)
 {
   rb_const_set(cCB_SUBST, ID_SUBST_INFO,
 	       allocate_cbsubst_info((struct cbsubst_info **)NULL));
 }
 
+static struct cbsubst_info *
+cbsubst_get_ptr(klass)
+    VALUE klass;
+{
+    return rb_check_typeddata(rb_const_get(klass, ID_SUBST_INFO), &cbsubst_info_type);
+}
+
 static VALUE
 cbsubst_initialize(argc, argv, self)
     int   argc;
@@ -1213,8 +1240,7 @@ cbsubst_initialize(argc, argv, self)
     struct cbsubst_info *inf;
     int idx, iv_idx;
 
-    Data_Get_Struct(rb_const_get(rb_obj_class(self), ID_SUBST_INFO),
-                    struct cbsubst_info, inf);
+    inf = cbsubst_get_ptr(rb_obj_class(self));
 
    idx = 0;
     for(iv_idx = 0; iv_idx < CBSUBST_TBL_MAX; iv_idx++) {
@@ -1281,12 +1307,11 @@ cbsubst_def_attr_aliases(self, tbl)
 {
     struct cbsubst_info *inf;
 
-    if (TYPE(tbl) != T_HASH) {
+    if (!RB_TYPE_P(tbl, T_HASH)) {
         rb_raise(rb_eArgError, "expected a Hash");
     }
 
-    Data_Get_Struct(rb_const_get(self, ID_SUBST_INFO),
-                    struct cbsubst_info, inf);
+    inf = cbsubst_get_ptr(self);
 
     rb_hash_foreach(tbl, each_attr_def, self);
 
@@ -1306,15 +1331,14 @@ cbsubst_sym_to_subst(self, sym)
     ID id;
     volatile VALUE ret;
 
-    if (TYPE(sym) != T_SYMBOL) return sym;
+    if (!RB_TYPE_P(sym, T_SYMBOL)) return sym;
 
-    Data_Get_Struct(rb_const_get(self, ID_SUBST_INFO),
-                    struct cbsubst_info, inf);
+    inf = cbsubst_get_ptr(self);
 
     if (!NIL_P(ret = rb_hash_aref(inf->aliases, sym))) {
-	str = rb_id2str(SYM2ID(ret));
+	str = rb_sym2str(ret);
     } else {
-	str = rb_id2str(SYM2ID(sym));
+	str = rb_sym2str(sym);
     }
 
     id = rb_intern_str(rb_sprintf("@%"PRIsVALUE, str));
@@ -1361,8 +1385,7 @@ cbsubst_get_subst_arg(argc, argv, self)
     ID id;
     volatile VALUE arg_sym, ret;
 
-    Data_Get_Struct(rb_const_get(self, ID_SUBST_INFO),
-                    struct cbsubst_info, inf);
+    inf = cbsubst_get_ptr(self);
 
     ptr = buf = ALLOC_N(char, inf->full_subst_length + 1);
 
@@ -1370,26 +1393,30 @@ cbsubst_get_subst_arg(argc, argv, self)
         switch(TYPE(argv[i])) {
         case T_STRING:
             str = argv[i];
-            arg_sym = ID2SYM(rb_intern_str(argv[i]));
+            arg_sym = rb_check_symbol(&str);
+            if (NIL_P(arg_sym)) goto not_found;
             break;
         case T_SYMBOL:
             arg_sym = argv[i];
-            str = rb_id2str(SYM2ID(arg_sym));
+            str = rb_sym2str(arg_sym);
             break;
         default:
             rb_raise(rb_eArgError, "arg #%d is not a String or a Symbol", i);
         }
 
         if (!NIL_P(ret = rb_hash_aref(inf->aliases, arg_sym))) {
-            str = rb_id2str(SYM2ID(ret));
+            str = rb_sym2str(ret);
         }
 
-        id = rb_intern_str(rb_sprintf("@%"PRIsVALUE, str));
+	ret = rb_sprintf("@%"PRIsVALUE, str);
+	id = rb_check_id(&ret);
+	if (!id) goto not_found;
 
 	for(idx = 0; idx < CBSUBST_TBL_MAX; idx++) {
 	  if (inf->ivar[idx] == id) break;
 	}
         if (idx >= CBSUBST_TBL_MAX) {
+	  not_found:
             rb_raise(rb_eArgError, "cannot find attribute :%"PRIsVALUE, str);
         }
 
@@ -1432,8 +1459,7 @@ cbsubst_get_subst_key(self, str)
     list = rb_funcall(cTclTkLib, ID_split_tklist, 1, str);
     len = RARRAY_LEN(list);
 
-    Data_Get_Struct(rb_const_get(self, ID_SUBST_INFO),
-                    struct cbsubst_info, inf);
+    inf = cbsubst_get_ptr(self);
 
     ptr = buf = ALLOC_N(char, inf->full_subst_length + len + 1);
 
@@ -1481,8 +1507,7 @@ cbsubst_get_all_subst_keys(self)
     long len;
     volatile VALUE ret;
 
-    Data_Get_Struct(rb_const_get(self, ID_SUBST_INFO),
-                    struct cbsubst_info, inf);
+    inf = cbsubst_get_ptr(self);
 
     ptr = buf = ALLOC_N(char, inf->full_subst_length + 1);
     keys_ptr = keys_buf = ALLOC_N(char, CBSUBST_TBL_MAX + 1);
@@ -1556,18 +1581,10 @@ cbsubst_table_setup(argc, argv, self)
   len = RARRAY_LEN(key_inf);
   for(idx = 0; idx < len; idx++) {
     inf = RARRAY_PTR(key_inf)[idx];
-    if (TYPE(inf) != T_ARRAY) continue;
+    if (!RB_TYPE_P(inf, T_ARRAY)) continue;
 
-    if (TYPE(RARRAY_PTR(inf)[0]) == T_STRING) {
-      chr = *(RSTRING_PTR(RARRAY_PTR(inf)[0]));
-    } else {
-      chr = NUM2CHR(RARRAY_PTR(inf)[0]);
-    }
-    if (TYPE(RARRAY_PTR(inf)[1]) == T_STRING) {
-      subst_inf->type[chr] = *(RSTRING_PTR(RARRAY_PTR(inf)[1]));
-    } else {
-      subst_inf->type[chr] = NUM2CHR(RARRAY_PTR(inf)[1]);
-    }
+    chr = NUM2CHR(RARRAY_PTR(inf)[0]);
+    subst_inf->type[chr] = NUM2CHR(RARRAY_PTR(inf)[1]);
 
     subst_inf->full_subst_length += 3;
 
@@ -1587,7 +1604,7 @@ cbsubst_table_setup(argc, argv, self)
   len = RARRAY_LEN(longkey_inf);
   for(idx = 0; idx < len; idx++) {
     inf = RARRAY_PTR(longkey_inf)[idx];
-    if (TYPE(inf) != T_ARRAY) continue;
+    if (!RB_TYPE_P(inf, T_ARRAY)) continue;
 
     chr = (unsigned char)(0x80 + idx);
     subst_inf->keylen[chr] = RSTRING_LEN(RARRAY_PTR(inf)[0]);
@@ -1602,11 +1619,7 @@ cbsubst_table_setup(argc, argv, self)
       subst_inf->key[chr][RSTRING_LEN(RARRAY_PTR(inf)[0])] = '\0';
     }
 #endif
-    if (TYPE(RARRAY_PTR(inf)[1]) == T_STRING) {
-      subst_inf->type[chr] = *(RSTRING_PTR(RARRAY_PTR(inf)[1]));
-    } else {
-      subst_inf->type[chr] = NUM2CHR(RARRAY_PTR(inf)[1]);
-    }
+    subst_inf->type[chr] = NUM2CHR(RARRAY_PTR(inf)[1]);
 
     subst_inf->full_subst_length += (subst_inf->keylen[chr] + 2);
 
@@ -1624,9 +1637,9 @@ cbsubst_table_setup(argc, argv, self)
   len = RARRAY_LEN(proc_inf);
   for(idx = 0; idx < len; idx++) {
     inf = RARRAY_PTR(proc_inf)[idx];
-    if (TYPE(inf) != T_ARRAY) continue;
+    if (!RB_TYPE_P(inf, T_ARRAY)) continue;
     rb_hash_aset(subst_inf->proc,
-		 ((TYPE(RARRAY_PTR(inf)[0]) == T_STRING)?
+		 (RB_TYPE_P(RARRAY_PTR(inf)[0], T_STRING)?
 		  INT2FIX(*(RSTRING_PTR(RARRAY_PTR(inf)[0]))) :
 		  RARRAY_PTR(inf)[0]),
 		 RARRAY_PTR(inf)[1]);
@@ -1666,8 +1679,7 @@ cbsubst_scan_args(self, arg_key, val_ary)
 
     old_gc = rb_gc_disable();
 
-    Data_Get_Struct(rb_const_get(self, ID_SUBST_INFO),
-                    struct cbsubst_info, inf);
+    inf = cbsubst_get_ptr(self);
 
     for(idx = 0; idx < vallen; idx++) {
       if (idx >= keylen) {
@@ -1734,7 +1746,7 @@ tkobj_path(self)
 const char tkutil_release_date[] = TKUTIL_RELEASE_DATE;
 
 void
-Init_tkutil()
+Init_tkutil(void)
 {
     VALUE cTK = rb_define_class("TkKernel", rb_cObject);
     VALUE mTK = rb_define_module("TkUtil");
@@ -1762,10 +1774,10 @@ Init_tkutil()
     ID_call = rb_intern("call");
 
     /* --------------------- */
-    cCB_SUBST = rb_define_class_under(mTK, "CallbackSubst", rb_cObject);
+    cCB_SUBST = rb_define_class_under(mTK, "CallbackSubst", rb_cData);
     rb_define_singleton_method(cCB_SUBST, "inspect", cbsubst_inspect, 0);
 
-    cSUBST_INFO = rb_define_class_under(cCB_SUBST, "Info", rb_cObject);
+    cSUBST_INFO = rb_define_class_under(cCB_SUBST, "Info", rb_cData);
     rb_define_singleton_method(cSUBST_INFO, "inspect", substinfo_inspect, 0);
 
     ID_SUBST_INFO = rb_intern("SUBST_INFO");
