@@ -11,13 +11,13 @@
 
 **********************************************************************/
 
-#include "ruby/ruby.h"
+#include "internal.h"
 #include "ruby/st.h"
 #include "ruby/util.h"
-#include "ruby/encoding.h"
-#include "internal.h"
 #include <errno.h>
 #include "probes.h"
+#include "id.h"
+#include "symbol.h"
 
 #ifdef __APPLE__
 # ifdef HAVE_CRT_EXTERNS_H
@@ -31,7 +31,6 @@
     ((klass = has_extra_methods(rb_obj_class(hash))) != 0) || \
     FL_TEST((hash), FL_EXIVAR|FL_TAINT|HASH_PROC_DEFAULT) || \
     !NIL_P(RHASH_IFNONE(hash)))
-#define HASH_REJECT_COPY_EXTRA_STATES 1
 
 static VALUE
 has_extra_methods(VALUE klass)
@@ -69,6 +68,12 @@ VALUE rb_cHash;
 static VALUE envtbl;
 static ID id_hash, id_yield, id_default, id_flatten_bang;
 
+VALUE
+rb_hash_ifnone(VALUE h)
+{
+    return RHASH_IFNONE(h);
+}
+
 VALUE
 rb_hash_set_ifnone(VALUE hash, VALUE ifnone)
 {
@@ -133,11 +138,26 @@ rb_any_hash(VALUE a)
 
     if (SPECIAL_CONST_P(a)) {
 	if (a == Qundef) return 0;
+	if (STATIC_SYM_P(a)) {
+	    a >>= (RUBY_SPECIAL_SHIFT + ID_SCOPE_SHIFT);
+	}
+	else if (FLONUM_P(a)) {
+	    /* prevent pathological behavior: [Bug #10761] */
+	    goto flt;
+	}
 	hnum = rb_objid_hash((st_index_t)a);
     }
     else if (BUILTIN_TYPE(a) == T_STRING) {
 	hnum = rb_str_hash(a);
     }
+    else if (BUILTIN_TYPE(a) == T_SYMBOL) {
+	return RSYMBOL(a)->hashval;
+    }
+    else if (BUILTIN_TYPE(a) == T_FLOAT) {
+      flt:
+	hval = rb_dbl_hash(rb_float_value(a));
+	hnum = FIX2LONG(hval);
+    }
     else {
         hval = rb_hash(a);
 	hnum = FIX2LONG(hval);
@@ -155,13 +175,50 @@ rb_objid_hash(st_index_t index)
     return hnum;
 }
 
+int
+rb_hash_iter_lev(VALUE h)
+{
+    return RHASH_ITER_LEV(h);
+}
+
 static const struct st_hash_type objhash = {
     rb_any_cmp,
     rb_any_hash,
 };
 
-extern const struct st_hash_type st_hashtype_num;
-#define identhash st_hashtype_num
+#define rb_ident_cmp st_numcmp
+
+static st_index_t
+rb_ident_hash(st_data_t n)
+{
+    /*
+     * This hash function is lightly-tuned for Ruby.  Further tuning
+     * should be possible.  Notes:
+     *
+     * - (n >> 3) alone is great for heap objects and OK for fixnum,
+     *   however symbols perform poorly.
+     * - (n >> (RUBY_SPECIAL_SHIFT+3)) was added to make symbols hash well,
+     *   n.b.: +3 to remove ID scope, +1 worked well initially, too
+     * - (n << 3) was finally added to avoid losing bits for fixnums
+     * - avoid expensive modulo instructions, it is currently only
+     *   shifts and bitmask operations.
+     * - flonum (on 64-bit) is pathologically bad, mix the actual
+     *   float value in, but do not use the float value as-is since
+     *   many integers get interpreted as 2.0 or -2.0 [Bug #10761]
+     */
+#ifdef USE_FLONUM /* RUBY */
+    if (FLONUM_P(n)) {
+	n ^= (st_data_t)rb_float_value(n);
+    }
+#endif
+
+    return (st_index_t)((n>>(RUBY_SPECIAL_SHIFT+3)|(n<<3)) ^ (n>>3));
+}
+
+static const struct st_hash_type identhash = {
+    rb_ident_cmp,
+    rb_ident_hash,
+};
 
 typedef int st_foreach_func(st_data_t, st_data_t, st_data_t);
 
@@ -300,7 +357,7 @@ rb_hash_new(void)
     return hash_alloc(rb_cHash);
 }
 
-static VALUE
+static inline VALUE
 rb_hash_dup_empty(VALUE hash)
 {
     NEWOBJ_OF(ret, struct RHash,
@@ -486,7 +543,8 @@ rb_hash_initialize(int argc, VALUE *argv, VALUE hash)
 	FL_SET(hash, HASH_PROC_DEFAULT);
     }
     else {
-	rb_scan_args(argc, argv, "01", &ifnone);
+	rb_check_arity(argc, 0, 1);
+	ifnone = argc == 0 ? Qnil : argv[0];
 	RHASH_SET_IFNONE(hash, ifnone);
     }
 
@@ -753,11 +811,12 @@ rb_hash_lookup(VALUE hash, VALUE key)
 static VALUE
 rb_hash_fetch_m(int argc, VALUE *argv, VALUE hash)
 {
-    VALUE key, if_none;
+    VALUE key;
     st_data_t val;
     long block_given;
 
-    rb_scan_args(argc, argv, "11", &key, &if_none);
+    rb_check_arity(argc, 1, 2);
+    key = argv[0];
 
     block_given = rb_block_given_p();
     if (block_given && argc == 2) {
@@ -773,7 +832,7 @@ rb_hash_fetch_m(int argc, VALUE *argv, VALUE hash)
 	    desc = rb_str_ellipsize(desc, 65);
 	    rb_raise(rb_eKeyError, "key not found: %"PRIsVALUE, desc);
 	}
-	return if_none;
+	return argv[1];
     }
     return (VALUE)val;
 }
@@ -810,7 +869,8 @@ rb_hash_default(int argc, VALUE *argv, VALUE hash)
 {
     VALUE key, ifnone;
 
-    rb_scan_args(argc, argv, "01", &key);
+    rb_check_arity(argc, 0, 1);
+    key = argv[0];
     ifnone = RHASH_IFNONE(hash);
     if (FL_TEST(hash, HASH_PROC_DEFAULT)) {
 	if (argc == 0) return Qnil;
@@ -885,7 +945,7 @@ rb_hash_default_proc(VALUE hash)
  *     h["cat"]   #=> "catcat"
  */
 
-static VALUE
+VALUE
 rb_hash_set_default_proc(VALUE hash, VALUE proc)
 {
     VALUE b;
@@ -956,22 +1016,48 @@ rb_hash_index(VALUE hash, VALUE value)
     return rb_hash_key(hash, value);
 }
 
-static VALUE
-rb_hash_delete_key(VALUE hash, VALUE key)
+/*
+ * delete a specified entry a given key.
+ * if there is the corresponding entry, return a value of the entry.
+ * if there is no corresponding entry, return Qundef.
+ */
+VALUE
+rb_hash_delete_entry(VALUE hash, VALUE key)
 {
     st_data_t ktmp = (st_data_t)key, val;
 
-    if (!RHASH(hash)->ntbl)
-        return Qundef;
-    if (RHASH_ITER_LEV(hash) > 0) {
-	if (st_delete_safe(RHASH(hash)->ntbl, &ktmp, &val, (st_data_t)Qundef)) {
-	    FL_SET(hash, HASH_DELETED);
-	    return (VALUE)val;
-	}
+    if (!RHASH(hash)->ntbl) {
+	return Qundef;
     }
-    else if (st_delete(RHASH(hash)->ntbl, &ktmp, &val))
+    else if (RHASH_ITER_LEV(hash) > 0 &&
+	     (st_delete_safe(RHASH(hash)->ntbl, &ktmp, &val, (st_data_t)Qundef))) {
+	FL_SET(hash, HASH_DELETED);
 	return (VALUE)val;
-    return Qundef;
+    }
+    else if (st_delete(RHASH(hash)->ntbl, &ktmp, &val)) {
+	return (VALUE)val;
+    }
+    else {
+	return Qundef;
+    }
+}
+
+/*
+ * delete a specified entry by a given key.
+ * if there is the corresponding entry, return a value of the entry.
+ * if there is no corresponding entry, return Qnil.
+ */
+VALUE
+rb_hash_delete(VALUE hash, VALUE key)
+{
+    VALUE deleted_value = rb_hash_delete_entry(hash, key);
+
+    if (deleted_value != Qundef) { /* likely pass */
+	return deleted_value;
+    }
+    else {
+	return Qnil;
+    }
 }
 
 /*
@@ -992,18 +1078,25 @@ rb_hash_delete_key(VALUE hash, VALUE key)
  *
  */
 
-VALUE
-rb_hash_delete(VALUE hash, VALUE key)
+static VALUE
+rb_hash_delete_m(VALUE hash, VALUE key)
 {
     VALUE val;
 
     rb_hash_modify_check(hash);
-    val = rb_hash_delete_key(hash, key);
-    if (val != Qundef) return val;
-    if (rb_block_given_p()) {
-	return rb_yield(key);
+    val = rb_hash_delete_entry(hash, key);
+
+    if (val != Qundef) {
+	return val;
+    }
+    else {
+	if (rb_block_given_p()) {
+	    return rb_yield(key);
+	}
+	else {
+	    return Qnil;
+	}
     }
-    return Qnil;
 }
 
 struct shift_var {
@@ -1050,7 +1143,7 @@ rb_hash_shift(VALUE hash)
 	else {
 	    rb_hash_foreach(hash, shift_i_safe, (VALUE)&var);
 	    if (var.key != Qundef) {
-		rb_hash_delete_key(hash, var.key);
+		rb_hash_delete_entry(hash, var.key);
 		return rb_assoc_new(var.key, var.val);
 	    }
 	}
@@ -1067,8 +1160,6 @@ delete_if_i(VALUE key, VALUE value, VALUE hash)
     return ST_CONTINUE;
 }
 
-static VALUE rb_hash_size(VALUE hash);
-
 static VALUE
 hash_enum_size(VALUE hash, VALUE args, VALUE eobj)
 {
@@ -1155,34 +1246,10 @@ rb_hash_reject(VALUE hash)
     if (RTEST(ruby_verbose)) {
 	VALUE klass;
 	if (HAS_EXTRA_STATES(hash, klass)) {
-#if HASH_REJECT_COPY_EXTRA_STATES
-	    rb_warn("copying extra states: %+"PRIsVALUE, hash);
-	    rb_warn("following states will not be copied in the future version:");
-	    if (klass) {
-		rb_warn("  subclass: %+"PRIsVALUE, klass);
-	    }
-	    if (FL_TEST(hash, FL_EXIVAR)) {
-		rb_warn("  instance variables: %+"PRIsVALUE,
-			rb_obj_instance_variables(hash));
-	    }
-	    if (FL_TEST(hash, FL_TAINT)) {
-		rb_warn("  taintedness");
-	    }
-	    if (FL_TEST(hash, HASH_PROC_DEFAULT)) {
-		rb_warn("  default proc: %+"PRIsVALUE, RHASH_IFNONE(hash));
-	    }
-	    else if (!NIL_P(RHASH_IFNONE(hash)))
-		rb_warn("  default value: %+"PRIsVALUE, RHASH_IFNONE(hash));
-#else
 	    rb_warn("extra states are no longer copied: %+"PRIsVALUE, hash);
-#endif
 	}
     }
-#if HASH_REJECT_COPY_EXTRA_STATES
-    result = rb_hash_dup_empty(hash);
-#else
     result = rb_hash_new();
-#endif
     if (!RHASH_EMPTY_P(hash)) {
 	rb_hash_foreach(hash, reject_i, result);
     }
@@ -1376,7 +1443,8 @@ NOINSERT_UPDATE_CALLBACK(hash_aset_str);
  *     h["a"] = 9
  *     h["c"] = 4
  *     h   #=> {"a"=>9, "b"=>200, "c"=>4}
- *     h.store("d", 42) #=> {"a"=>9, "b"=>200, "c"=>4, "d"=>42}
+ *     h.store("d", 42) #=> 42
+ *     h   #=> {"a"=>9, "b"=>200, "c"=>4, "d"=>42}
  *
  *  +key+ should not have its value changed while it is in use as a key (an
  *  <tt>unfrozen String</tt> passed as a key will be duplicated and frozen).
@@ -1501,7 +1569,7 @@ rb_hash_replace(VALUE hash, VALUE hash2)
  *     h.length        #=> 3
  */
 
-static VALUE
+VALUE
 rb_hash_size(VALUE hash)
 {
     return INT2FIX(RHASH_SIZE(hash));
@@ -1791,7 +1859,7 @@ rb_hash_keys(VALUE hash)
     if (ST_DATA_COMPATIBLE_P(VALUE)) {
 	st_table *table = RHASH(hash)->ntbl;
 
-	if (OBJ_PROMOTED(keys)) rb_gc_writebarrier_remember_promoted(keys);
+	rb_gc_writebarrier_remember(keys);
 	RARRAY_PTR_USE(keys, ptr, {
 	    size = st_keys_check(table, ptr, size, Qundef);
 	});
@@ -1835,7 +1903,7 @@ rb_hash_values(VALUE hash)
     if (ST_DATA_COMPATIBLE_P(VALUE)) {
 	st_table *table = RHASH(hash)->ntbl;
 
-	if (OBJ_PROMOTED(values)) rb_gc_writebarrier_remember_promoted(values);
+	rb_gc_writebarrier_remember(values);
 	RARRAY_PTR_USE(values, ptr, {
 	    size = st_values_check(table, ptr, size, Qundef);
 	});
@@ -1863,7 +1931,7 @@ rb_hash_values(VALUE hash)
  *
  */
 
-static VALUE
+VALUE
 rb_hash_has_key(VALUE hash, VALUE key)
 {
     if (!RHASH(hash)->ntbl)
@@ -1953,13 +2021,20 @@ hash_equal(VALUE hash1, VALUE hash2, int eql)
 
     if (hash1 == hash2) return Qtrue;
     if (!RB_TYPE_P(hash2, T_HASH)) {
-	if (!rb_respond_to(hash2, rb_intern("to_hash"))) {
+	if (!rb_respond_to(hash2, idTo_hash)) {
 	    return Qfalse;
 	}
-	if (eql)
-	    return rb_eql(hash2, hash1);
-	else
+	if (eql) {
+	    if (rb_eql(hash2, hash1)) {
+		return Qtrue;
+	    }
+	    else {
+		return Qfalse;
+	    }
+	}
+	else {
 	    return rb_equal(hash2, hash1);
+	}
     }
     if (RHASH_SIZE(hash1) != RHASH_SIZE(hash2))
 	return Qfalse;
@@ -2035,6 +2110,8 @@ hash_i(VALUE key, VALUE val, VALUE arg)
  *
  *  Compute a hash-code for this hash. Two hashes with the same content
  *  will have the same hash code (and will compare using <code>eql?</code>).
+ *
+ *  See also Object#hash.
  */
 
 static VALUE
@@ -2436,7 +2513,7 @@ static VALUE rb_hash_compare_by_id_p(VALUE hash);
  *     h1["a"]        #=> 100
  *     h1.compare_by_identity
  *     h1.compare_by_identity? #=> true
- *     h1["a"]        #=> nil  # different objects.
+ *     h1["a".dup]    #=> nil  # different objects.
  *     h1[:c]         #=> "c"  # same symbols are all same.
  *
  */
@@ -2471,6 +2548,72 @@ rb_hash_compare_by_id_p(VALUE hash)
     return Qfalse;
 }
 
+VALUE
+rb_ident_hash_new(void)
+{
+    VALUE hash = rb_hash_new();
+    RHASH(hash)->ntbl = st_init_table(&identhash);
+    return hash;
+}
+
+st_table *
+rb_init_identtable(void)
+{
+    return st_init_table(&identhash);
+}
+
+st_table *
+rb_init_identtable_with_size(st_index_t size)
+{
+    return st_init_table_with_size(&identhash, size);
+}
+
+static int
+any_p_i(VALUE key, VALUE value, VALUE arg)
+{
+    VALUE ret = rb_yield(rb_assoc_new(key, value));
+    if (RTEST(ret)) {
+	*(VALUE *)arg = Qtrue;
+	return ST_STOP;
+    }
+    return ST_CONTINUE;
+}
+
+static int
+any_p_i_fast(VALUE key, VALUE value, VALUE arg)
+{
+    VALUE ret = rb_yield_values(2, key, value);
+    if (RTEST(ret)) {
+	*(VALUE *)arg = Qtrue;
+	return ST_STOP;
+    }
+    return ST_CONTINUE;
+}
+
+/*
+ *  call-seq:
+ *     hsh.any? [{ |(key, value)| block }]   -> true or false
+ *
+ *  See also Enumerable#any?
+ */
+
+static VALUE
+rb_hash_any_p(VALUE hash)
+{
+    VALUE ret = Qfalse;
+
+    if (RHASH_EMPTY_P(hash)) return Qfalse;
+    if (!rb_block_given_p()) {
+	/* yields pairs, never false */
+	return Qtrue;
+    }
+    if (rb_block_arity() > 1)
+	rb_hash_foreach(hash, any_p_i_fast, (VALUE)&ret);
+    else
+	rb_hash_foreach(hash, any_p_i, (VALUE)&ret);
+    return ret;
+}
+
 static int path_tainted = -1;
 
 static char **origenviron;
@@ -2511,11 +2654,20 @@ extern char **environ;
 #define ENVNMATCH(s1, s2, n) (memcmp((s1), (s2), (n)) == 0)
 #endif
 
+#ifdef _WIN32
+static VALUE
+env_str_transcode(VALUE str, rb_encoding *enc)
+{
+    return rb_str_conv_enc_opts(str, rb_utf8_encoding(), enc,
+				ECONV_INVALID_REPLACE | ECONV_UNDEF_REPLACE, Qnil);
+}
+#endif
+
 static VALUE
 env_str_new(const char *ptr, long len)
 {
 #ifdef _WIN32
-    VALUE str = rb_str_conv_enc(rb_str_new(ptr, len), rb_utf8_encoding(), rb_locale_encoding());
+    VALUE str = env_str_transcode(rb_str_new(ptr, len), rb_locale_encoding());
 #else
     VALUE str = rb_locale_str_new(ptr, len);
 #endif
@@ -2524,6 +2676,19 @@ env_str_new(const char *ptr, long len)
     return str;
 }
 
+static VALUE
+env_path_str_new(const char *ptr)
+{
+#ifdef _WIN32
+    VALUE str = env_str_transcode(rb_str_new_cstr(ptr), rb_filesystem_encoding());
+#else
+    VALUE str = rb_filesystem_str_new_cstr(ptr);
+#endif
+
+    rb_obj_freeze(str);
+    return str;
+}
+
 static VALUE
 env_str_new2(const char *ptr)
 {
@@ -2531,16 +2696,27 @@ env_str_new2(const char *ptr)
     return env_str_new(ptr, strlen(ptr));
 }
 
+#define get_env_ptr(var, val) \
+    (memchr((var = RSTRING_PTR(val)), '\0', RSTRING_LEN(val)) ? \
+     rb_raise(rb_eArgError, "bad environment variable " #var) : (void)0)
+
+static inline const char *
+env_name(volatile VALUE *s)
+{
+    const char *name;
+    SafeStringValue(*s);
+    get_env_ptr(name, *s);
+    return name;
+}
+
+#define env_name(s) env_name(&(s))
+
 static VALUE
 env_delete(VALUE obj, VALUE name)
 {
-    char *nam, *val;
+    const char *nam, *val;
 
-    SafeStringValue(name);
-    nam = RSTRING_PTR(name);
-    if (memchr(nam, '\0', RSTRING_LEN(name))) {
-	rb_raise(rb_eArgError, "bad environment variable name");
-    }
+    nam = env_name(name);
     val = getenv(nam);
     if (val) {
 	VALUE value = env_str_new2(val);
@@ -2585,24 +2761,13 @@ static int env_path_tainted(const char *);
 static VALUE
 rb_f_getenv(VALUE obj, VALUE name)
 {
-    char *nam, *env;
+    const char *nam, *env;
 
-    SafeStringValue(name);
-    nam = RSTRING_PTR(name);
-    if (memchr(nam, '\0', RSTRING_LEN(name))) {
-	rb_raise(rb_eArgError, "bad environment variable name");
-    }
+    nam = env_name(name);
     env = getenv(nam);
     if (env) {
 	if (ENVMATCH(nam, PATH_ENV) && !env_path_tainted(env)) {
-#ifdef _WIN32
-	    VALUE str = rb_str_conv_enc(rb_str_new(env, strlen(env)), rb_utf8_encoding(), rb_filesystem_encoding());
-#else
-	    VALUE str = rb_filesystem_str_new_cstr(env);
-#endif
-
-	    rb_obj_freeze(str);
-	    return str;
+	    return env_path_str_new(env);
 	}
 	return env_str_new2(env);
     }
@@ -2626,34 +2791,27 @@ rb_f_getenv(VALUE obj, VALUE name)
 static VALUE
 env_fetch(int argc, VALUE *argv)
 {
-    VALUE key, if_none;
+    VALUE key;
     long block_given;
-    char *nam, *env;
+    const char *nam, *env;
 
-    rb_scan_args(argc, argv, "11", &key, &if_none);
+    rb_check_arity(argc, 1, 2);
+    key = argv[0];
     block_given = rb_block_given_p();
     if (block_given && argc == 2) {
 	rb_warn("block supersedes default value argument");
     }
-    SafeStringValue(key);
-    nam = RSTRING_PTR(key);
-    if (memchr(nam, '\0', RSTRING_LEN(key))) {
-	rb_raise(rb_eArgError, "bad environment variable name");
-    }
+    nam = env_name(key);
     env = getenv(nam);
     if (!env) {
 	if (block_given) return rb_yield(key);
 	if (argc == 1) {
 	    rb_raise(rb_eKeyError, "key not found: \"%"PRIsVALUE"\"", key);
 	}
-	return if_none;
+	return argv[1];
     }
     if (ENVMATCH(nam, PATH_ENV) && !env_path_tainted(env))
-#ifdef _WIN32
-	return rb_str_conv_enc(rb_str_new(env, strlen(env)), rb_utf8_encoding(), rb_filesystem_encoding());
-#else
-	return rb_filesystem_str_new_cstr(env);
-#endif
+	return env_path_str_new(env);
     return env_str_new2(env);
 }
 
@@ -2718,13 +2876,15 @@ getenvsize(const char* p)
     return p - porg + 1;
 }
 static size_t
-getenvblocksize()
+getenvblocksize(void)
 {
     return (rb_w32_osver() >= 5) ? 32767 : 5120;
 }
 #endif
 
-#if !defined(HAVE_SETENV) || !defined(HAVE_UNSETENV)
+#if defined(_WIN32) || \
+  (defined(__sun) && !(defined(HAVE_SETENV) && defined(HAVE_UNSETENV)))
+
 NORETURN(static void invalid_envname(const char *name));
 
 static void
@@ -2793,10 +2953,22 @@ ruby_setenv(const char *name, const char *value)
 #endif
     }
 #elif defined __sun
-    size_t len;
-    char **env_ptr, *str;
+    /* Solaris 9 (or earlier) does not have setenv(3C) and unsetenv(3C). */
+    /* The below code was tested on Solaris 10 by:
+         % ./configure ac_cv_func_setenv=no ac_cv_func_unsetenv=no
+    */
+    size_t len, mem_size;
+    char **env_ptr, *str, *mem_ptr;
 
+    check_envname(name);
     len = strlen(name);
+    if (value) {
+	mem_size = len + strlen(value) + 2;
+	mem_ptr = malloc(mem_size);
+	if (mem_ptr == NULL)
+	    rb_sys_fail_str(rb_sprintf("malloc("PRIuSIZE")", mem_size));
+	snprintf(mem_ptr, mem_size, "%s=%s", name, value);
+    }
     for (env_ptr = GET_ENVIRON(environ); (str = *env_ptr) != 0; ++env_ptr) {
 	if (!strncmp(str, name, len) && str[len] == '=') {
 	    if (!in_origenv(str)) free(str);
@@ -2805,10 +2977,10 @@ ruby_setenv(const char *name, const char *value)
 	}
     }
     if (value) {
-	str = malloc(len += strlen(value) + 2);
-	snprintf(str, len, "%s=%s", name, value);
-	if (putenv(str))
+	if (putenv(mem_ptr)) {
+	    free(mem_ptr);
 	    rb_sys_fail_str(rb_sprintf("putenv(%s)", name));
+	}
     }
 #else  /* WIN32 */
     size_t len;
@@ -2878,12 +3050,10 @@ env_aset(VALUE obj, VALUE nm, VALUE val)
     }
     SafeStringValue(nm);
     SafeStringValue(val);
-    name = RSTRING_PTR(nm);
-    value = RSTRING_PTR(val);
-    if (memchr(name, '\0', RSTRING_LEN(nm)))
-	rb_raise(rb_eArgError, "bad environment variable name");
-    if (memchr(value, '\0', RSTRING_LEN(val)))
-	rb_raise(rb_eArgError, "bad environment variable value");
+    /* nm can be modified in `val.to_str`, don't get `name` before
+     * check for `val` */
+    get_env_ptr(name, nm);
+    get_env_ptr(value, val);
 
     ruby_setenv(name, value);
     if (ENVMATCH(name, PATH_ENV)) {
@@ -3370,12 +3540,9 @@ env_empty_p(void)
 static VALUE
 env_has_key(VALUE env, VALUE key)
 {
-    char *s;
+    const char *s;
 
-    SafeStringValue(key);
-    s = RSTRING_PTR(key);
-    if (memchr(s, '\0', RSTRING_LEN(key)))
-	rb_raise(rb_eArgError, "bad environment variable name");
+    s = env_name(key);
     if (getenv(s)) return Qtrue;
     return Qfalse;
 }
@@ -3390,12 +3557,9 @@ env_has_key(VALUE env, VALUE key)
 static VALUE
 env_assoc(VALUE env, VALUE key)
 {
-    char *s, *e;
+    const char *s, *e;
 
-    SafeStringValue(key);
-    s = RSTRING_PTR(key);
-    if (memchr(s, '\0', RSTRING_LEN(key)))
-	rb_raise(rb_eArgError, "bad environment variable name");
+    s = env_name(key);
     e = getenv(s);
     if (e) return rb_assoc_new(key, rb_tainted_str_new2(e));
     return Qnil;
@@ -3821,7 +3985,7 @@ Init_Hash(void)
     rb_define_method(rb_cHash,"values_at", rb_hash_values_at, -1);
 
     rb_define_method(rb_cHash,"shift", rb_hash_shift, 0);
-    rb_define_method(rb_cHash,"delete", rb_hash_delete, 1);
+    rb_define_method(rb_cHash,"delete", rb_hash_delete_m, 1);
     rb_define_method(rb_cHash,"delete_if", rb_hash_delete_if, 0);
     rb_define_method(rb_cHash,"keep_if", rb_hash_keep_if, 0);
     rb_define_method(rb_cHash,"select", rb_hash_select, 0);
@@ -3848,6 +4012,8 @@ Init_Hash(void)
     rb_define_method(rb_cHash,"compare_by_identity", rb_hash_compare_by_id, 0);
     rb_define_method(rb_cHash,"compare_by_identity?", rb_hash_compare_by_id_p, 0);
 
+    rb_define_method(rb_cHash, "any?", rb_hash_any_p, 0);
+
     /* Document-class: ENV
      *
      * ENV is a hash-like accessor for environment variables.
