@@ -2,14 +2,13 @@
 
   inits.c -
 
-  $Author: akr $
+  $Author$
   created at: Tue Dec 28 16:01:58 JST 1993
 
   Copyright (C) 1993-2007 Yukihiro Matsumoto
 
 **********************************************************************/
 
-#include "ruby/ruby.h"
 #include "internal.h"
 
 #define CALL(n) {void Init_##n(void); Init_##n();}
@@ -17,6 +16,7 @@
 void
 rb_call_inits(void)
 {
+    CALL(Method);
     CALL(RandomSeed);
     CALL(sym);
     CALL(var_tables);
@@ -46,7 +46,6 @@ rb_call_inits(void)
     CALL(Time);
     CALL(Random);
     CALL(signal);
-    CALL(process);
     CALL(load);
     CALL(Proc);
     CALL(Binding);
@@ -56,9 +55,11 @@ rb_call_inits(void)
     CALL(VM);
     CALL(ISeq);
     CALL(Thread);
+    CALL(process);
     CALL(Cont);
     CALL(Rational);
     CALL(Complex);
     CALL(version);
+    CALL(vm_trace);
 }
 #undef CALL
