@@ -59,7 +59,6 @@ The original copyright notice follows.
    email: matumoto@math.keio.ac.jp
 */
 
-#include "ruby/ruby.h"
 #include "internal.h"
 
 #include <limits.h>
@@ -335,7 +334,7 @@ static const rb_data_type_t random_data_type = {
 	random_free,
 	random_memsize,
     },
-    NULL, NULL, RUBY_TYPED_FREE_IMMEDIATELY
+    0, 0, RUBY_TYPED_FREE_IMMEDIATELY
 };
 
 static rb_random_t *
@@ -474,14 +473,14 @@ fill_random_seed(uint32_t seed[DEFAULT_SEED_CNT])
 
     gettimeofday(&tv, 0);
     seed[0] ^= tv.tv_usec;
-    seed[1] ^= (unsigned int)tv.tv_sec;
+    seed[1] ^= (uint32_t)tv.tv_sec;
 #if SIZEOF_TIME_T > SIZEOF_INT
-    seed[0] ^= (unsigned int)((time_t)tv.tv_sec >> SIZEOF_INT * CHAR_BIT);
+    seed[0] ^= (uint32_t)((time_t)tv.tv_sec >> SIZEOF_INT * CHAR_BIT);
 #endif
     seed[2] ^= getpid() ^ (n++ << 16);
-    seed[3] ^= (unsigned int)(VALUE)&seed;
+    seed[3] ^= (uint32_t)(VALUE)&seed;
 #if SIZEOF_VOIDP > SIZEOF_INT
-    seed[2] ^= (unsigned int)((VALUE)&seed >> SIZEOF_INT * CHAR_BIT);
+    seed[2] ^= (uint32_t)((VALUE)&seed >> SIZEOF_INT * CHAR_BIT);
 #endif
 }
 
@@ -938,7 +937,7 @@ rand_int(struct MT *mt, VALUE vmax, int restrictive)
     else {
 	VALUE ret;
 	if (rb_bigzero_p(vmax)) return Qnil;
-	if (!RBIGNUM_SIGN(vmax)) {
+	if (!BIGNUM_SIGN(vmax)) {
 	    if (restrictive) return Qnil;
             vmax = rb_big_uminus(vmax);
 	}
@@ -985,7 +984,7 @@ rand_range(struct MT* mt, VALUE range)
 		v = ULONG2NUM(r);
 	    }
 	}
-	else if (BUILTIN_TYPE(vmax) == T_BIGNUM && RBIGNUM_SIGN(vmax) && !rb_bigzero_p(vmax)) {
+	else if (BUILTIN_TYPE(vmax) == T_BIGNUM && BIGNUM_SIGN(vmax) && !rb_bigzero_p(vmax)) {
 	    vmax = excl ? rb_big_minus(vmax, INT2FIX(1)) : rb_big_norm(vmax);
 	    if (FIXNUM_P(vmax)) {
 		excl = 0;
@@ -1352,7 +1351,7 @@ rb_reset_random_seed(void)
  */
 
 void
-Init_Random(void)
+InitVM_Random(void)
 {
     Init_RandomSeed2();
     rb_define_global_function("srand", rb_f_srand, -1);
@@ -1383,7 +1382,14 @@ Init_Random(void)
     rb_define_singleton_method(rb_cRandom, "new_seed", random_seed, 0);
     rb_define_private_method(CLASS_OF(rb_cRandom), "state", random_s_state, 0);
     rb_define_private_method(CLASS_OF(rb_cRandom), "left", random_s_left, 0);
+}
 
+#undef rb_intern
+void
+Init_Random(void)
+{
     id_rand = rb_intern("rand");
     id_bytes = rb_intern("bytes");
+
+    InitVM(Random);
 }
