@@ -3,7 +3,6 @@ require 'test/unit'
 require 'net/http'
 require 'stringio'
 require_relative 'utils'
-require_relative '../../ruby/envutil'
 
 class TestNetHTTP < Test::Unit::TestCase
 
@@ -140,12 +139,12 @@ class TestNetHTTP < Test::Unit::TestCase
 
   def test_proxy_port
     clean_http_proxy_env do
-      http = Net::HTTP.new 'example', nil, 'proxy.example'
+      http = Net::HTTP.new 'exmaple', nil, 'proxy.example'
       assert_equal 'proxy.example', http.proxy_address
       assert_equal 80, http.proxy_port
-      http = Net::HTTP.new 'example', nil, 'proxy.example', 8000
+      http = Net::HTTP.new 'exmaple', nil, 'proxy.example', 8000
       assert_equal 8000, http.proxy_port
-      http = Net::HTTP.new 'example', nil
+      http = Net::HTTP.new 'exmaple', nil
       assert_equal nil, http.proxy_port
     end
   end
@@ -277,6 +276,7 @@ module TestNetHTTP_version_1_1_methods
       end
     }
     assert_equal 1, i
+    @log_tester = nil # server may encount ECONNRESET
   end
 
   def test_get__implicit_start
@@ -397,6 +397,7 @@ module TestNetHTTP_version_1_1_methods
   def test_timeout_during_HTTP_session
     bug4246 = "expected the HTTP session to have timed out but have not. c.f. [ruby-core:34203]"
 
+    th = nil
     # listen for connections... but deliberately do not read
     TCPServer.open('localhost', 0) {|server|
       port = server.addr[1]
@@ -412,6 +413,9 @@ module TestNetHTTP_version_1_1_methods
       end
       assert th.join(10), bug4246
     }
+  ensure
+    th.kill
+    th.join
   end
 end
 
@@ -559,10 +563,10 @@ module TestNetHTTP_version_1_2_methods
   end
 
   def _test_request__uri_host(http)
-    uri = URI 'http://example/'
+    uri = URI 'http://other.example/'
 
     req = Net::HTTP::Get.new(uri)
-    req['host'] = 'other.example'
+    req['host'] = 'example'
 
     res = http.request(req)
 
@@ -822,22 +826,6 @@ class TestNetHTTPContinue < Test::Unit::TestCase
     assert_not_match(/HTTP\/1.1 100 continue/, @debug.string)
   end
 
-  def test_expect_continue_error_before_body
-    @log_tester = nil
-    mount_proc {|req, res|
-      raise WEBrick::HTTPStatus::Forbidden
-    }
-    start {|http|
-      uheader = {'content-length' => '5', 'expect' => '100-continue'}
-      http.continue_timeout = 1 # allow the server to respond before sending
-      http.request_post('/continue', 'data', uheader) {|res|
-        assert_equal(res.code, '403')
-      }
-    }
-    assert_match(/Expect: 100-continue/, @debug.string)
-    assert_not_match(/HTTP\/1.1 100 continue/, @debug.string)
-  end
-
   def test_expect_continue_error_while_waiting
     mount_proc {|req, res|
       res.status = 501
