@@ -1,6 +1,5 @@
 require 'test/unit'
 require 'pp'
-require_relative 'envutil'
 
 $m0 = Module.nesting
 
@@ -678,9 +677,18 @@ class TestModule < Test::Unit::TestCase
 
   def test_const_set_invalid_name
     c1 = Class.new
-    assert_raise(NameError) { c1.const_set(:foo, :foo) }
-    assert_raise(NameError) { c1.const_set("bar", :foo) }
-    assert_raise(TypeError) { c1.const_set(1, :foo) }
+    assert_raise_with_message(NameError, /foo/) { c1.const_set(:foo, :foo) }
+    assert_raise_with_message(NameError, /bar/) { c1.const_set("bar", :foo) }
+    assert_raise_with_message(TypeError, /1/) { c1.const_set(1, :foo) }
+    assert_nothing_raised(NameError) { c1.const_set("X\u{3042}", :foo) }
+    assert_raise(NameError) { c1.const_set("X\u{3042}".encode("utf-16be"), :foo) }
+    assert_raise(NameError) { c1.const_set("X\u{3042}".encode("utf-16le"), :foo) }
+    assert_raise(NameError) { c1.const_set("X\u{3042}".encode("utf-32be"), :foo) }
+    assert_raise(NameError) { c1.const_set("X\u{3042}".encode("utf-32le"), :foo) }
+    cx = EnvUtil.labeled_class("X\u{3042}")
+    EnvUtil.with_default_internal(Encoding::UTF_8) {
+      assert_raise_with_message(TypeError, /X\u{3042}/) { c1.const_set(cx, :foo) }
+    }
   end
 
   def test_const_get_invalid_name
@@ -909,25 +917,34 @@ class TestModule < Test::Unit::TestCase
     assert_include(c.constants(false), :Foo, bug9413)
   end
 
-  def test_frozen_class
+  def test_frozen_module
     m = Module.new
     m.freeze
     assert_raise(RuntimeError) do
       m.instance_eval { undef_method(:foo) }
     end
+  end
 
+  def test_frozen_class
     c = Class.new
     c.freeze
     assert_raise(RuntimeError) do
       c.instance_eval { undef_method(:foo) }
     end
+  end
 
-    o = Object.new
+  def test_frozen_singleton_class
+    klass = Class.new
+    o = klass.new
     c = class << o; self; end
     c.freeze
-    assert_raise(RuntimeError) do
+    assert_raise_with_message(RuntimeError, /frozen/) do
       c.instance_eval { undef_method(:foo) }
     end
+    klass.class_eval do
+      def self.foo
+      end
+    end
   end
 
   def test_method_defined
@@ -1272,7 +1289,7 @@ class TestModule < Test::Unit::TestCase
     c = Class.new do
       attr_writer :foo
     end
-    assert_raise(ArgumentError) { c.new.send :foo= }
+    assert_raise(ArgumentError, bug8540) { c.new.send :foo= }
   end
 
   def test_private_constant
@@ -1629,7 +1646,7 @@ class TestModule < Test::Unit::TestCase
 
   def test_prepend_visibility_inherited
     bug8238 = '[ruby-core:54105] [Bug #8238]'
-    assert_separately [], <<-"end;", timeout: 3
+    assert_separately [], <<-"end;", timeout: 20
       class A
         def foo() A; end
         private :foo
@@ -1708,6 +1725,17 @@ class TestModule < Test::Unit::TestCase
     assert_equal('hello!', foo.new.hello, bug9236)
   end
 
+  def test_prepend_call_super
+    assert_separately([], <<-'end;') #do
+      bug10847 = '[ruby-core:68093] [Bug #10847]'
+      module M; end
+      Float.prepend M
+      assert_nothing_raised(SystemStackError, bug10847) do
+        0.3.numerator
+      end
+    end;
+  end
+
   def test_class_variables
     m = Module.new
     m.class_variable_set(:@@foo, 1)
@@ -1812,6 +1840,14 @@ class TestModule < Test::Unit::TestCase
     assert_warning '' do
       assert_equal(42, a.ivar)
     end
+
+    name = "@\u{5909 6570}"
+    assert_warning(/instance variable #{name} not initialized/) do
+      val = EnvUtil.with_default_external(Encoding::UTF_8) {
+        a.instance_eval(name)
+      }
+      assert_nil(val)
+    end
   end
 
   def test_uninitialized_attr
