@@ -2,7 +2,7 @@
 
   version.c -
 
-  $Author: nobu $
+  $Author$
   created at: Thu Sep 30 20:08:01 JST 1993
 
   Copyright (C) 1993-2007 Yukihiro Matsumoto
@@ -78,6 +78,9 @@ void
 ruby_show_version(void)
 {
     PRINT(description);
+#ifdef HAVE_MALLOC_CONF
+    if (malloc_conf) printf("malloc_conf=%s\n", malloc_conf);
+#endif
     fflush(stdout);
 }
 
